
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <propeller2.h>

enum {
	// _xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 4_000_000,
	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,


	SPI_CE_PIN = 32,
	SPI_CLK_PIN = 33,
	SPI_TX_PIN = 34,
	SPI_RX_PIN = 35,

	SPI_CPOL = 0,   // clock polarity, as per SPI (set for modes 2 and 3)
	SPI_CPHA = 0,   // clock phase, as per SPI (set for modes 1 and 3)
	CLK_DIV = 8,    // sysclock divider for SPI clock gen
	CLK_ADV = 0,    // advance/retard (sysclocks) of tx clock/data phase relationship
	CLK_REGD = 0,   //    clock pin:  0 for unregistered, 1 for registered
	TX_REGD = 1,    // tx data pins:  0 for unregistered, 1 for registered

	CMD_WRITE1 = 0x02,
	CMD_FREAD1 = 0x0b,
	CMD_RESET_EN  = 0x66,
	CMD_RESET     = 0x99,
	CMD_READ_ID   = 0x9f,
};


static void spi_shift_out( uint8_t *paddr, uint32_t len )
{
	uint32_t  mdat = X_RFBYTE_1P_1DAC1 | SPI_TX_PIN<<17 | X_PINS_ON | X_ALT_ON | len<<3;  // SPI_DI is first pin

	__asm volatile { // no optimising and enforces Fcache use - Needed to free up the FIFO
		rdfast  xfrq, paddr         // non-blocking FIFO prefetch from haddr (byte granular)
		setxfrq xfrq                // set sysclock/1 for lead in timing
		shr     paddr, #1
		movbyts paddr, #0x1b        // big-endian
		setbyte paddr, #CMD_WRITE1, #0    // insert Command into CA word
		shl     len, #3
		add	len, #32            // number of SPI clock cycles

		drvl    #SPI_TX_PIN         // drive tx pin for CA phase
		drvl    #SPI_CE_PIN

		xinit   mleadin, #0         // lead-in timing, at sysclock/1
		setq    mnco                // streamer transfer rate
		xcont   mca, paddr          // tx Command and Address (byte granular address)
		dirh    #SPI_CLK_PIN        // start smartpin internally cycling at SPI clock rate
		wypin   len, #SPI_CLK_PIN   // SPI clocks for CA phase and data phase

		xcont   mdat, #0            // tx data from FIFO

		// Possible spare time here for other maintanence.  But this is CE low time, so don't over do it.

		waitxfi                     // wait until streamer is done
		dirl    #SPI_CLK_PIN        // reset smartpin, just the right number of clock pulses
		dirl    #SPI_TX_PIN         // tristate the databus upon completion
	_ret_	drvh    #SPI_CE_PIN

xfrq		long    0x8000_0000
mleadin		long    X_IMM_32X1_1DAC1 | (5 + CLK_ADV + CLK_DIV + CLK_REGD - TX_REGD - (1 ^ SPI_CPHA) * (CLK_DIV>>1))
mnco		long    0x8000_0000UL / CLK_DIV + (0x8000_0000UL % CLK_DIV > 0 ? 1 : 0)    // round up
mca		long    X_IMM_32X1_1DAC1 | SPI_TX_PIN<<17 | X_PINS_ON | X_ALT_ON | 32  // SPI_DI is first pin
	} // Exiting of Fcache/Cogexec issues an implicit RDFAST, flushing remaining FIFO content to hubRAM
}


static void  init_smartpins()
{
    _pinh( SPI_CE_PIN );
    _wrpin( SPI_CE_PIN, P_SYNC_IO );

    _pinf( SPI_TX_PIN );
    _wrpin( SPI_TX_PIN, TX_REGD ? P_SYNC_IO : 0 );  // at sysclock/1, data pins should be registered to minimise relative skew

    _pinf( SPI_CLK_PIN );
    _wrpin( SPI_CLK_PIN, P_PULSE | P_OE | (CLK_REGD ? P_SYNC_IO : 0) | (SPI_CPOL ? P_INVERT_OUTPUT : 0) );  // clock pins for CA phase
    _wxpin( SPI_CLK_PIN, CLK_DIV | (CLK_DIV>>1 + (SPI_CPOL & CLK_DIV))<<16 );    // set period and duty
}


uint8_t  txstack[1024];
uint8_t  txdata[4096];


void  spi_master( int some )
{
    for( int i=0; i<400; i++ )
        txdata[i] = 400 - i;

    _waitms( 100 );
    init_smartpins();

    spi_shift_out( txdata, 400 );
    _waitms( 100 );
}



enum {
	CS = 32,
	CLK = 33,
	MOSI = 35,
	MISO = 34,
	CSMask = 1U << CS,
	CLKMask = 1U << CLK,
	MOSIMask = 1U << MOSI,
	MISOMask = 1U << MISO,
};


static int spi_shift_in(unsigned char *addr)
{
	int data, count = 0;    // bytes received

	__asm volatile { // no optimising and enforces Fcache use - Needed to free up the FIFO
		fltl	#MOSI    // disable/reset SPI RX
.wait1
		testp	#CS   wc
	if_nc	jmp	#.wait1    // wait while CS low

		wrfast	#0, addr    // setup the hubRAM FIFO hardware for sequencial writes
		dirh	#MOSI    // enable smart pin SPI RX
.wait2
		testp	#CS   wc
	if_c	jmp	#.wait2    // wait while CS high

		rep	@.rend, #0    // loop forever

		testp	#CS   wz
		testp	#MOSI   wc
if_nc_and_z	jmp	#.rend    // break loop if CS high
	if_c	rdpin	data, #MOSI   // collect the received byte
	if_c	rev	data
	if_c	wfbyte	data    // write to FIFO (hubRAM)
	if_c	add	count, #1
.rend

		dirl	#MOSI    // disable/reset SPI RX
	} // Exiting of Fcache/Cogexec issues an implicit RDFAST, flushing remaining FIFO content to hubRAM

    return count;
}


uint8_t  rxdata[4096];


void main(void)
{
    int m, x, ret;

    _waitms(200);
    printf( "\n   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode() );

    _cogstart_C( spi_master, 0, txstack, 1024 );

    _pinclear(CS);
    m = P_SYNC_RX;
    m |= ((MISO-MOSI) & 0b111) << 28;  // smartA (rx pin) input select
    m |= ((CLK-MOSI) & 0b111) << 24;  // smartB (clock pin) input select
    x = 0b0_00000 | (8-1);  // early sample alignment and 8-bit words
    _pinstart(MOSI, m, x, 0);

    while(1){
            ret = 0;
            ret = spi_shift_in( rxdata );
            if(ret>0)
                    printf( " Byte count = %d\n", ret );
                    for(int i = 0; i<ret; i++){
                         printf("%02x ",rxdata[i]);
                    }
                    printf( "\n" );
    }
}
